
/* dither.c */
extern int dither(MAP *, int, int, int, int, int, const char *);

/* main.c */
extern void view_message(const char *, const char *, ...);
extern PALETTE *alloc_palette(void);
extern void alloc_image(MAP *, int, int);
extern void free_image(MAP *);
extern MAP *alloc_map(void);
extern void free_map(MAP *);
extern void status_bar(int, int);
extern int main(int, char **);

/* read-degas.c */
extern int maybe_read_degas(FILE *, const char *, MAP *);

/* read-gif.c */
extern int maybe_read_gif(FILE *, const char *, MAP *);

/* read-pnm.c */
extern int maybe_read_pnm(FILE *, const char *, MAP *);

/* read-uxf.c */
extern int maybe_read_uxface(FILE *, const char *, MAP *);

/* read.c */
extern MAP *read_map(FILE *, const char *);

/* view-vga.c */
extern int readkey(void);
extern void plot_pixel(int, int, COLOR);
extern int init_display(int *, int *, int *, const char *);
extern int set_colors(PALETTE *);
extern void tidy_display(void);
